
// FEATURE:   <code type="c++"/>          -- syntax highlighting
// FEATURE:   <code linenumbers="true"/>  -- LaTeX moreverb package can help
// FIXME: nest TextNodes within each other for bold+italic

package org.ibex.util;
import java.util.*;
import java.io.*;
import org.ibex.util.*;

public class Doc extends XML {

    Root root = null;
    int skip = 0;
    Vec nodeStack = new Vec();
    String pending = "";
    char[] buffer = new char[1024 * 1024];
    int preStart = -1;
    
    public Doc() { }

    public static void main(String[] s) throws Exception {
        Doc d = new Doc();
        Reader r = new InputStreamReader(System.in);
        int len = 0;
        while(true) {
            int numread = r.read(d.buffer, len, d.buffer.length - len);
            if (numread == -1) break;
            len += numread;
            if (len >= d.buffer.length) {
                char[] newbuffer = new char[d.buffer.length * 2];
                System.arraycopy(d.buffer, 0, newbuffer, 0, d.buffer.length);
                d.buffer = newbuffer;
            }
        }
        d.parse(new CharArrayReader(d.buffer));
        StringBuffer sb = new StringBuffer();
        d.root.dumpLatex(sb);
        System.out.println(sb);
    }

    public void startElement(Element e) throws Exn {
        if (preStart != -1) return;
        String name = e.getLocalName();
        Node newGuy = null;
        if (name.equals("ibex-doc")) {            newGuy = new Root(e);
        } else if (name.equals("section")) {      newGuy = new Section(e);
        } else if (name.equals("heading")) {      newGuy = new Heading(e);
        } else if (name.equals("appendix")) {     newGuy = new Appendix(e);
        } else if (name.equals("list")) {         newGuy = new List();

        } else if (name.equals("pre")) {          preStart = getGlobalOffset();

        } else if (name.equals("definition")) {   newGuy = new Definition(e);
        } else if (name.equals("property")) {     newGuy = new Property(e);

        } else if (name.equals("b")) {            newGuy = new B();
        } else if (name.equals("i")) {            newGuy = new I();
        } else if (name.equals("t")) {            newGuy = new T();
        } else if (name.equals("link")) {         newGuy = new Link(e);
        } else if (name.equals("image")) {        newGuy = new Image(e);

        } else {
            System.err.println("warning: unknown tag " + name);
            skip++;
        }
        if (newGuy != null) {
            Node parent = (Node)nodeStack.lastElement();
            if (parent != null) parent.addChild(newGuy);
            nodeStack.addElement(newGuy);
        }
    }

    public void whitespace(char[] ch, int start, int length) throws Exn, IOException { characters(ch, start, length); }
    public void endElement(Element e) throws Exn, IOException {
        if (preStart != -1) {
            if (!e.getLocalName().equals("pre")) return;
            ((Node)nodeStack.lastElement()).addChild(new PRE(preStart, getGlobalOffset()));
            preStart = -1;
        } else if (skip > 0) {
            skip--;
            return;
        } else {
            nodeStack.setSize(nodeStack.size() - 1);
        }
    }
    public void characters(char[] ch, int start, int length) throws Exn, IOException {
        if (preStart != -1) return;
        ((Node)nodeStack.lastElement()).addText(new String(ch, start, length));
    }

    abstract class Node {
        public Node parent = null;
        public abstract void addText(String s);
        public abstract void addChild(Node s);
        public abstract void dumpLatex(StringBuffer sb);
        final String fixLatex(String s) {
            if (s == null) return "";
            s = s.replaceAll("\\$", "\\\\\\$ ");
            s = s.replaceAll("\\\\([^\\$])", "\\$\\\\backslash\\$\\1");
            s = s.replaceAll("LaTeX", "\\\\LaTeX");
            s = s.replaceAll("\\%", "\\\\% ");
            s = s.replaceAll("#", "\\\\#");
            s = s.replaceAll("\\{", "\\\\{");
            s = s.replaceAll("\\}", "\\\\}");
            s = s.replaceAll("\\&", "\\\\&");
            s = s.replaceAll("\\~", "\\\\~");
            s = s.replaceAll("_", "\\\\_");
            if (!(this instanceof T)) {
                s = s.replaceAll(" \"", " ``");
                s = s.replaceAll("\"", "''");
            }
            return s;
        }
    }


    // Empty Nodes //////////////////////////////////////////////////////////////////////////////

    class EmptyNode extends Node {
        public EmptyNode() { }
        public void addChild(Node o) { throw new RuntimeException(this.getClass().getName() + " cannot have children"); }
        public void addText(String o) { throw new RuntimeException(this.getClass().getName() + " cannot have content"); }
        public void dumpLatex(StringBuffer sb) { }
    }

    class Image extends EmptyNode {
        public String url;
        public Image(XML.Element e) { url = e.getAttrVal("url"); }
        public void dumpLatex(StringBuffer sb) { sb.append("\\hyperimage{" + url + "}"); }
    }

    class Heading extends EmptyNode {
        public String text;
        public Heading(XML.Element e) { text = e.getAttrVal("title"); }
        public void dumpLatex(StringBuffer sb) { sb.append("\\vspace{.6cm}\\hypertarget{" +
                                                           fixLatex(text) + "}{\\textbf{\\textsf{" +
                                                           fixLatex(text) + "}}}"); }
    }

    class LineBreak extends EmptyNode {
        public LineBreak() { }
        public void dumpLatex(StringBuffer sb) { sb.append("\\\n"); }
    }

    class ParagraphBreak extends EmptyNode {
        public ParagraphBreak() { }
        public void dumpLatex(StringBuffer sb) { sb.append("\n\n"); }
    }



    // Non-Paragraph Nodes //////////////////////////////////////////////////////////////////////////////

    /** Nodes which contain only text; they split themselves if anything else is added */
    class TextNode extends Node implements Cloneable {
        protected String mytext = "";
        private boolean canAcceptMoreText = true;
        public void dumpLatex(StringBuffer sb) { sb.append(fixLatex(mytext)); }
        public void addChild(Node o) { canAcceptMoreText = false; parent.addChild(o); }
        public void addText(String o) {
            if (canAcceptMoreText) {
                mytext += (String)o;
                String[] split = mytext.split("\\n\\s*\\n");
                if (split.length <= 1) return;
                canAcceptMoreText = false;
                mytext = split[0];
                parent.addChild(new ParagraphBreak());
                o = "";
                for(int i=1; i<split.length; i++) o += split[i];
            }
            try {
                TextNode clone = (TextNode)clone();
                clone.mytext = "";
                clone.canAcceptMoreText = true;
                clone.addText(o);
                parent.addChild(clone);
            } catch (CloneNotSupportedException cnse) {
                throw new RuntimeException(cnse);
            }
        }
    }

    class T extends TextNode {
        public void dumpLatex(StringBuffer sb) {
            sb.append("{\\texttt{");
            super.dumpLatex(sb);
            sb.append("}}");
        } }
    class B extends TextNode {
        public void dumpLatex(StringBuffer sb) {sb.append("{\\textbf{"); super.dumpLatex(sb); sb.append("}}");}}
    class I extends TextNode {
        public void dumpLatex(StringBuffer sb) {sb.append("{\\it{"); super.dumpLatex(sb); sb.append("}}");}}
    class Link extends TextNode {
        public String url;
        public String section;
        public String appendix;
        public String text;
        public Link(XML.Element e) {
            url = e.getAttrVal("url");
            text = e.getAttrVal("text");
            appendix = e.getAttrVal("appendix");
            section = e.getAttrVal("section");
        }
        public void dumpLatex(StringBuffer sb) {
            // FIXME: dotted underline for section/appendix, solid underline for url
            String text = fixLatex(this.text);
            if (text == null) text = "\\tt " + (url != null ? url : section != null ? section : appendix);
            if (url != null) {             sb.append("\\href{" + url + "}{\\uline{" + text + "}} ");
            } else if (section != null) {  sb.append("\\hyperlink{" + section + "}{\\uwave{" + text + "}} ");
            } else if (appendix != null) { sb.append("\\hyperlink{" + appendix + "}{\\uwave{" + text + "}} ");
            }
        }
    }

    class PRE extends TextNode {
        public PRE(int start, int end) {
            while(Character.isWhitespace(buffer[start])) start++;
            while(buffer[start] != '\n') start--;
            start++;
            end -= 6; // ugly hack since </pre> is 6 chars long...
            if (Character.isWhitespace(buffer[end])) {
                while(Character.isWhitespace(buffer[end])) end--;
                end++;
            }
            mytext = new String(buffer, start, end - start);
        }
        public void dumpLatex(StringBuffer sb) {
            sb.append("\n\n\\begin{Verbatim}[fontfamily=courier,fontsize=\\footnotesize,frame=single,rulecolor=\\color{CodeBorder},resetmargins=true]\n");
            sb.append(mytext);
            sb.append("\\end{Verbatim}\n\n");
        }
    }


    // Paragraph Nodes //////////////////////////////////////////////////////////////////////////////

    class ParagraphNode extends Node {
        protected Vec children = new Vec();
        public void addChild(Node n) { children.addElement(n); n.parent = this; }
        public void addText(String s) {
            if (children.size() == 0 || !(children.lastElement() instanceof String))
                children.addElement(s);
            else
                children.setElementAt(children.lastElement() + s, children.size() - 1);
        }
        public void dumpLatex(StringBuffer sb) {
            for(int i=0; i<children.size(); i++) {
                if (children.elementAt(i) instanceof String) sb.append(fixLatex(children.elementAt(i).toString()));
                else ((Node)children.elementAt(i)).dumpLatex(sb);
            }
        }
    }

    class Property extends ParagraphNode {
        String name = "";
        String type = "";
        String default_ = null;
        public Property(XML.Element e) {
            name = e.getAttrVal("name");
            if (name == null || name.equals("")) name = "ERROR";
            type = e.getAttrVal("type");
            default_ = e.getAttrVal("default");
            if (default_ != null && default_.trim().length() == 0) default_ = null;
        }
        public void dumpLatex(StringBuffer sb) {
            StringBuffer sb2 = new StringBuffer();
            super.dumpLatex(sb2);
            String s = sb2.toString();
            while(Character.isSpace(s.charAt(0))) s = s.substring(1);
            String fname = fixLatex(name);
            type = type == null || type.equals("") ? "" : "({\\it{" + fixLatex(type) + "}})";
            if (name.trim().length() > 13) {
                sb.append("\n\n{\\color{CodeBorder}\\hspace{-2cm}\\dotfill\\\\\\color{black}}");
                sb.append("\\marginpar{\\raggedleft{\\texttt{\\textbf{\\footnotesize{"+fname+"}}}}\\\\"+type+" }");
                sb.append("\\\\");
                sb.append(s);
            } else {
                sb.append("\n\n{\\color{CodeBorder}\\hspace{-2cm}\\dotfill\\\\\\color{black}}");
                sb.append(s.substring(0, s.indexOf(' ')));
                sb.append("\\marginpar{\\raggedleft{\\texttt{\\textbf{\\footnotesize{"+fname+"}}}}\\\\"+type+" }");
                sb.append(s.substring(s.indexOf(' ')));
            }
            if (default_ != null) {
                sb.append("\\\\{\\it default: }{\\texttt{" + fixLatex(default_) + "}}\n\n");
            } else {
                sb.append("\n\n");
            }
        }
    }

    class Definition extends ParagraphNode {
        String name = "";
        public Definition(XML.Element e) { name = e.getAttrVal("term"); }
        public void dumpLatex(StringBuffer sb) {
            StringBuffer sb2 = new StringBuffer();
            super.dumpLatex(sb2);
            String s = sb2.toString();
            while(Character.isSpace(s.charAt(0))) s = s.substring(1);
            sb.append(s.substring(0, s.indexOf(' ')));
            sb.append("\\marginpar{\\raggedleft{\\textbf{"+fixLatex(name)+"}}}");
            sb.append(s.substring(s.indexOf(' ')));
        }
    }

    class Section extends ParagraphNode {
        String name;
        public Section(XML.Element e) { name = e.getAttrVal("title"); }
        public void dumpLatex(StringBuffer sb) {
            String secs = "";
            String base = "section";
            for(Node n = parent; n != null; n = n.parent)
                if (n instanceof Section || n instanceof Appendix) {
                    base = "section";
                    secs += "sub";
                }
            if (secs.length() == 0)
                sb.append("\\newpage\n\n");
            sb.append("\n\n\\hypertarget{" + name + "}{\\" + secs + base + "{" + name + "}}\n\n");
            super.dumpLatex(sb);
        }
    }

    static boolean hitAppendix = false;
    class Appendix extends Section {
        public Appendix(XML.Element e) { super(e); }
        public void dumpLatex(StringBuffer sb) {
            if (!hitAppendix) sb.append("\\appendix\n");
            hitAppendix = true;
            super.dumpLatex(sb);
        }
    }

    class List extends ParagraphNode {
        Node textnode = null;
        public List() { }
        public void addText(String s) {
            if (!(children.lastElement() == textnode) || (textnode == null && children.lastElement() == null))
                addChild(textnode = new TextNode());
            ((TextNode)children.lastElement()).addText(s);
        }
        public void dumpLatex(StringBuffer sb) {
            sb.append("\n\\begin{itemize}%\n");
            boolean unusedItem = false;
            for(int i=0; i<children.size(); i++) {
                Object kid = children.elementAt(i);
                if (kid instanceof ParagraphBreak || kid instanceof List) unusedItem = false;
                if (kid instanceof String) {
                    if (kid.toString().trim().length() == 0) continue;
                    if (!unusedItem) { unusedItem = true; sb.append("\n\n\\item\n"); }
                    sb.append(children.elementAt(i));
                } else {
                    if (kid instanceof TextNode) {
                        if (((TextNode)kid).mytext.trim().length() == 0) continue;
                        if (!unusedItem) { unusedItem = true; sb.append("\n\n\\item\n"); }
                    }
                    ((Node)kid).dumpLatex(sb);
                }
            }
            sb.append("\n\\end{itemize}%\n");
        }
    }

    class Root extends ParagraphNode {
        String title = "You forgot the title, you idiot!";
        String author = "Your Mom";
        String email = null;
        String subtitle = null;
        public Root(XML.Element e) {
            root = this;
            title = e.getAttrVal("title");
            author = e.getAttrVal("author");
            email = e.getAttrVal("email");
            subtitle = e.getAttrVal("subtitle");
        }
        public void dumpLatex(StringBuffer sb) {
            sb.append("\\documentclass{article}\n");
            sb.append("\\def\\ninept{\\def\\baselinestretch{.95}\\let\\normalsize\\small\\normalsize}\n");
            sb.append("\\ninept\n");
            sb.append("\\usepackage{color}\n");
            sb.append("\\definecolor{CodeBorder}{rgb}{0.6,0.6,0.6}\n");
            sb.append("\\definecolor{CodeBackground}{rgb}{0.93,0.93,0.93}\n");
            sb.append("\\usepackage{graphicx}\n");
            sb.append("\\usepackage{courier}\n");
            sb.append("\\usepackage{fancyvrb}\n");
            sb.append("\\usepackage{fvrb-ex}\n");
            sb.append("\\usepackage{bold-extra}\n");
            sb.append("\\usepackage{ulem}\n");
            sb.append("\\usepackage{appendix}\n");
            sb.append("\\usepackage{amssymb,amsmath,epsfig,alltt}\n");
            sb.append("\\sloppy\n");
            sb.append("\\usepackage{palatino}\n");
            sb.append("\\usepackage{sectsty}\n");
            sb.append("\\allsectionsfont{\\sffamily}\n");
            sb.append("\\sectionfont{\\color{black}\\leftskip=-2cm\\hrulefill\\\\\\sffamily\\bfseries\\raggedleft\\vspace{1cm}}\n");
            sb.append("\\subsectionfont{\\color{black}\\dotfill\\\\\\sffamily\\raggedright\\hspace{-4cm}}\n");
            sb.append("\\newdimen\\sectskip\n");
            sb.append("\\newdimen\\subsectskip\n");
            sb.append("\\newdimen\\saveskip\n");
            sb.append("\\saveskip=\\leftskip\n");
            sb.append("\\sectskip=-2cm\n");
            sb.append("\\subsectskip=0cm\n");
            sb.append("\\let\\oldsection\\section\n");
            sb.append("\\let\\oldsubsection\\subsection\n");
            sb.append("\\def\\subsection#1{\\leftskip=\\sectskip\\oldsubsection{#1}\\leftskip=0cm}\n");
            sb.append("\\usepackage{parskip}\n");
            sb.append("\\usepackage{tabularx}\n");
            sb.append("\\usepackage{alltt}\n");
            sb.append("\\usepackage[pdftex,colorlinks=true,urlcolor=blue,linkcolor=blue,bookmarks=true]{hyperref}\n");
            // FIXME: pdfauthor, pdftitle, pdfsubject, pdfkeywords?
            sb.append("\n");
            sb.append("\\begin{document}\n");
            sb.append("\\reversemarginpar\n");
            sb.append("\n");
            sb.append("\\title{\\textbf{\\textsf{\n");
            sb.append(title);
            if (subtitle != null) sb.append("\\\\{\\large " + subtitle + "}\n");
            sb.append("}}}\n");
            if (author != null) {
                sb.append("\\author{\n");
                sb.append(author);
                if (email != null) sb.append("\\\\{\\tt " + email + "}\n");
                sb.append("}\n");
            }
            sb.append("\n");
            sb.append("\\maketitle\n");
            sb.append("\\clearpage\n");
            sb.append("\\tableofcontents\n");
            sb.append("\\clearpage\n");
            sb.append("\\onecolumn\n");
            super.dumpLatex(sb);
            sb.append("\\end{document}");
        }
    }
}

